import os
from itertools import tee, izip
import tarfile
import ctypes

import numpy as np
from scipy.ndimage.filters import maximum_filter, minimum_filter
from scipy.interpolate import RectSphereBivariateSpline

from c_wrapper import cextrema

EARTH_RADIUS = 6371


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return izip(a, b)


def find_extrema_slow(array, print_warning=True):
    '''
    Takes an array and finds its local extrema.

    Returns an array with 0s for not an extrema, 1s for maxs and -1 for mins
    and a list of the indices of all maximums and minimums
    '''
    if print_warning:
        print('Warning, this function is quite slow. Consider using find_extrema(array) instead')
    extrema = np.zeros_like(array)
    maximums = []
    minimums = []
    for i in range(1, array.shape[0] - 1):
        for j in range(0, array.shape[1]):
            val = array[i, j]

            is_max, is_min = True, True
            for ii in range(i - 1, i + 2):
                for jj in range(j - 1, j + 2):
                    if val < array[ii, jj % array.shape[1]]:
                        is_max = False
                    elif val > array[ii, jj % array.shape[1]]:
                        is_min = False
            if is_max:
                extrema[i, j] = 1
                maximums.append((i, j))
            elif is_min:
                extrema[i, j] = -1
                minimums.append((i, j))
    return extrema, maximums, minimums


def find_extrema(array):
    '''
    Takes an array and finds its local extrema.

    Returns an array with 0s for not an extrema, 1s for maxs and -1 for mins
    and a list of the indices of all maximums and minimums

    N.B. this function is much faster than the above.
    '''
    extrema = np.zeros_like(array)
    maximums = []
    minimums = []

    local_max = maximum_filter(array, size=(3, 3)) == array
    local_min = minimum_filter(array, size=(3, 3)) == array
    extrema += local_max
    extrema -= local_min

    where_max = np.where(local_max)
    where_min = np.where(local_min)

    for max_point in zip(where_max[0], where_max[1]):
        if (max_point[0] != 0 and max_point[0] != array.shape[0] - 1 and
            max_point[1] != 0 and max_point[1] != array.shape[1] - 1):
            maximums.append(max_point)

    for min_point in zip(where_min[0], where_min[1]):
        if (min_point[0] != 0 and min_point[0] != array.shape[0] - 1 and
            min_point[1] != 0 and min_point[1] != array.shape[1] - 1):
            minimums.append(min_point)

    return extrema, maximums, minimums


MAX_MAX_MINS = 1000
def cfind_extrema(array):
    extrema = np.zeros_like(array)

    max_x = np.zeros(MAX_MAX_MINS, dtype=np.int32)
    max_y = np.zeros(MAX_MAX_MINS, dtype=np.int32)
    min_x = np.zeros(MAX_MAX_MINS, dtype=np.int32)
    min_y = np.zeros(MAX_MAX_MINS, dtype=np.int32)

    max_length = ctypes.c_int();
    min_length = ctypes.c_int();

    cextrema(array, array.shape[0], array.shape[1], extrema, MAX_MAX_MINS, max_x, max_y, min_x,
            min_y, ctypes.byref(max_length), ctypes.byref(min_length))

    return extrema, zip(max_x[:max_length.value], max_y[:max_length.value]), zip(min_x[:min_length.value], min_y[:min_length.value])


def upscale_field(lons, lats, field, x_scale=2, y_scale=2, is_degrees=True):
    '''
    Takes a field defined on a sphere using lons/lats and returns an upscaled
    version, using cubic spline interpolation.
    '''
    if is_degrees:
        lons = lons * np.pi / 180.
        lats = (lats + 90) * np.pi / 180.

    d_lon = lons[1] - lons[0]
    d_lat = lats[1] - lats[0]

    new_lon = np.linspace(lons[0], lons[-1], len(lons) * x_scale)
    new_lat = np.linspace(lats[0], lats[-1], len(lats) * x_scale)

    mesh_new_lat, mesh_new_lon = np.meshgrid(new_lat, new_lon)

    if True:
        lut = RectSphereBivariateSpline(lats[1:-1], lons[1:-1], field[1:-1, 1:-1])

        interp_field = lut.ev(mesh_new_lat[1:-1, 1:-1].ravel(),
                              mesh_new_lon[1:-1, 1:-1].ravel()).reshape(mesh_new_lon.shape[0] - 2,
                                                                        mesh_new_lon.shape[1] - 2).T
    else:
        pass
    if is_degrees:
        new_lon = new_lon * 180. / np.pi
        new_lat = (new_lat * 180. / np.pi) - 90

    return new_lon[1:-1], new_lat[1:-1], interp_field


def geo_dist(p1, p2):
    '''Returns the geodesic distance between two points

    p1, p2 should be of the form (lon, lat) in degrees
    '''
    lon1, lat1 = np.pi * p1[0] / 180., np.pi * p1[1] / 180.
    lon2, lat2 = np.pi * p2[0] / 180., np.pi * p2[1] / 180.
    return np.arccos(
        np.sin(lat1) * np.sin(lat2) +
        np.cos(lat1) * np.cos(lat2) * np.cos(lon2 - lon1)) * EARTH_RADIUS


def dist(p1, p2):
    '''Returns the cartesian distance between two points'''
    return ((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2) ** 0.5


def raster_voronoi(extrema, maximums, minimums):
    '''
    Takes a 2D array and points of max/mins, and returns a 2D array
    with the voronoi sections filled with different values
    '''
    voronoi_arr = np.zeros_like(extrema)
    for i in range(1, voronoi_arr.shape[0] - 1):
        for j in range(0, voronoi_arr.shape[1]):
            min_dist = 1e9
            for k, extrema_point in enumerate(minimums + maximums):
                test_dist = dist((i, j), extrema_point)
                if test_dist < min_dist:
                    min_dist = test_dist
                    voronoi_arr[i, j] = k

    for k, extrema_point in enumerate(minimums + maximums):
        voronoi[extrema_point[0], extrema_point[1]] = 0
    voronoi[voronoi > len(minimums)] = -1

    return voronoi


def compress_file(data_file):
    curr_dir = os.getcwd()
    parent_dir = os.path.dirname(data_file)
    os.chdir(parent_dir)

    compressed_filename = data_file + '.bz2'
    tar = tarfile.open(compressed_filename, 'w:bz2')
    tar.add(os.path.basename(data_file))
    tar.close()

    os.chdir(curr_dir)
    return compressed_filename


def compress_dir(data_dir):
    '''Compresses a given data_dir, taking care to get file names correct'''
    curr_dir = os.getcwd()

    parent_dir = os.path.dirname(data_dir)
    os.chdir(parent_dir)
    compressed_filename = data_dir + '.bz2'
    tar = tarfile.open(compressed_filename, 'w:bz2')
    for root, dirs, files in os.walk(data_dir):
        for file in files:
            tar.add(os.path.relpath(os.path.join(root, file), start=parent_dir))
    tar.close()

    os.chdir(curr_dir)
    return compressed_filename


def decompress_file(compressed_filename):
    '''Decompresses a given tarball, taking care to get file names correct'''
    tar = tarfile.open(compressed_filename)
    tar.extractall(os.path.dirname(compressed_filename))
    tar.close()



import os
import time
import datetime as dt

import numpy as np
from netCDF4 import Dataset
from scipy.interpolate import interp1d
import scipy.ndimage as ndimage

from utils.c_wrapper import cvort, cvort4
from utils.utils import cfind_extrema, upscale_field
from load_settings import settings
import setup_logging

C20_DATA_DIR = os.path.join(settings.DATA_DIR, 'c20_full')

EARTH_RADIUS = 6371000
EARTH_CIRC = EARTH_RADIUS * 2 * np.pi
NUM_ENSEMBLE_MEMBERS = 56

log = setup_logging.get_logger('st.find_vortmax')


class C20Data(object):
    '''Class used for accessing data from C20 Reanalysis project.

    This acts as a wrapper around netCDF4.Datasets and makes it easy to view data.
    Typically it exposes the prmsl and vort850/vort9950 fields for all ensemble members.
    It will load these fields, along with corresponding maxima (vorticity) and minima (pressure)
    each time a new date is set.

    :param year: Year from which to take data
    :param fields: List of C20 fields that are to be loaded, or use 'all' for complete set
    :param version: Version of C20 data to use
    '''

    def __init__(self, year, fields='all', version=settings.C20_VERSION):
        self._year = year
        self.dx = None
        self.date = None
	self.version = version
	log.info('C20Data: year={}, version={}'.format(year, version))

        if fields == 'all':
            # rh995 has been removed.
            self.fields = ['u9950', 'v9950', 'u850', 'v850', 'prmsl', 't9950', 't850', 'cape', 'pwat']
        else:
            self.fields = fields

        if 'u9950' in self.fields and 'v9950' in self.fields:
            self.calc_9950_vorticity = True
        else:
            self.calc_9950_vorticity = False

        if 'u850' in self.fields and 'v850' in self.fields:
            self.calc_850_vorticity = True
        else:
            self.calc_850_vorticity = False

        fields = ', '.join(self.fields)
	log.info('Using: {}'.format(fields))
        self._load_datasets(self._year)

    def set_year(self, year):
        '''Sets a year and loads the relevant dataset'''
        self._year = year
        self.close_datasets()
        self._load_datasets(self._year)

    def close_datasets(self):
        '''Closes all open datasets'''
        for dataset in self.nc_datasets.values():
            dataset.close()

    def _load_datasets(self, year):
        '''Loads datasets for a given year

        Just sets up the NetCDF4 objects, doesn't actually load any data apart from
        lons/lats and dates.
        '''
        # All datasets have lon/lat/time info in them, so any will do.
        any_dataset = None
        dataset_fieldname = None
        self.nc_datasets = {}

        for field in self.fields:
            # e.g. ~/stormtracks_data/data/c20_full/2005/prmsl_2005.nc
	    path = os.path.join(C20_DATA_DIR, self.version, str(year), '{}_{}.nc'.format(field, year))
	    if not os.path.exists(path):
		msg = 'File does not exist: {}'.format(path)
		log.error(msg)
		raise RuntimeError(msg)
	    log.debug('Loading {} from {}'.format(field, path))
            dataset = Dataset(path)
            dataset_fieldname = field
            any_dataset = dataset
            self.nc_datasets[field] = dataset

        start_date = dt.datetime(1, 1, 1)
        hours_since_JC = any_dataset.variables['time'][:]
        self.number_enseble_members = any_dataset.variables[dataset_fieldname].shape[1]

        self.lons = any_dataset.variables['lon'][:]
        self.lats = any_dataset.variables['lat'][:]

        self.dates = np.array([start_date + dt.timedelta(hs / 24.) -
                              dt.timedelta(2) for hs in hours_since_JC])

        dlon = self.lons[2] - self.lons[0]

        # N.B. array as dx varies with lat.
        # lons, lats are in degres.
        self.dx = (dlon * np.cos(self.lats * np.pi / 180) * EARTH_CIRC) / 360.
        self.dy = (self.lats[0] - self.lats[2]) * EARTH_CIRC / 360.

        # Interpolation functions.
        self.f_lon = interp1d(np.arange(0, 180), self.lons)
        self.f_lat = interp1d(np.arange(0, 91), self.lats)
	self.first_date()

    def first_date(self):
        '''Sets date to the first date of the year (i.e. Jan the 1st)'''
        return self.set_date(self.dates[0])

    def next_date(self):
        '''Moves date on by one timestep (6hr)'''
        index = np.where(self.dates == self.date)[0][0]
        if index < len(self.dates):
            date = self.dates[index + 1]
            return self.set_date(date)
        else:
	    log.warn('Trying to set date beyond date range')
            return None

    def prev_date(self):
        '''Moves date back by one timestep (6hr)'''
        index = np.where(self.dates == self.date)[0][0]
        if index > 0:
            date = self.dates[index - 1]
            return self.set_date(date)
        else:
	    log.warn('Trying to set date beyond date range')
            return None

    def set_date(self, date):
        '''Sets date and loads all data for that date

        Will have no effect if there is no difference in date.

        :param date: date to load

        :returns: date if successful, otherwise None
        '''
        if date != self.date:
            try:
                log.debug("Setting date to {0}".format(date))
                index = np.where(self.dates == date)[0][0]
                self.date = date
                self._process_ensemble_data(index)
            except:
                self.date = None
		log.exception('Problem loading date {}'.format(date))
                raise
        return date

    def _cvorticity(self, u, v):
        '''Calculates the (2nd order) vorticity by calling into a c function'''
        vort = np.zeros_like(u)
        cvort(u, v, u.shape[0], u.shape[1], self.dx, self.dy, vort)
        return vort

    def _cvorticity4(self, u, v):
        '''Calculates the (4th order) vorticity by calling into a c function

        Algorithm was taken from Walsh's code'''
        vort = np.zeros_like(u)
        cvort4(u, v, u.shape[0], u.shape[1], self.dx, self.dy, vort)
        return vort

    def _process_ensemble_data(self, index):
        '''
        Processes data for one ensemble member

        Loads the relevant data and then performs a variety of calculations on it.
        At a minimum, prmsl, vort and vort4 will be calculated for the current date, as well
        as their maxima/minima as appropriate. Additionally (depending on how class is configured),
        smoothed_vort and up_vort (upscaled_vorticity) can be calculated.

        Rough times for each step are recorded.

        :param index: index of timestep in C20 data
        '''
        start = time.time()
        self._load_ensemble_data(index)
        end = time.time()
        fields = ', '.join(self.fields)
        log.debug('  Loaded {0} in {1}'.format(fields, end - start))

        if self.calc_9950_vorticity:
            start = time.time()
            self._calculate_vorticities('9950')
            end = time.time()
            log.debug('  Calculated 9950 vorticity in {0}'.format(end - start))
        if self.calc_850_vorticity:
            start = time.time()
            self._calculate_vorticities('850')
            end = time.time()
            log.debug('  Calculated 850 vorticity in {0}'.format(end - start))

        start = time.time()
        self._find_min_max_from_fields()
        end = time.time()
        log.debug('  Found maxima/minima in {0}'.format(end - start))

    def _load_ensemble_data(self, index):
        '''Loads the raw data from the NetCDF4 files'''
        # N.B. it is very important how the data is loaded. The data is stored in NetCDF4 files,
        # which in turn uses HDF5 as a storage medium. HDF5 allows for compression of particular
        # subsets of data ('chunks'). If you access the data in terms of these chunks, it will be
        # **much** faster, which is why all data for one date is loaded at a time, i.e. 56x91x180
        # cells, or num_ensemble_members x lat x lon.
        # This can be seen by looking at e.g. c20data.prmsl.shape, which will be (56, 91, 180).
        for field in self.fields:
            if field in ['u9950', 'u850', 'u250']:
                setattr(self, field, - self.nc_datasets[field].variables[field][index])
            else:
                setattr(self, field, self.nc_datasets[field].variables[field][index])

    def _calculate_vorticities(self, pressure_level):
        '''Calculates vort (2nd order) and vort4 (4th order)

        Uses c functions for speed.'''
        vort = []
        # self.vort4 = []
        if pressure_level == '9950':
            for em in range(NUM_ENSEMBLE_MEMBERS):
                vort.append(self._cvorticity(self.u9950[em], self.v9950[em]))
                # vort4.append(self._cvorticity4(self.u[em], self.v[em]))
        elif pressure_level == '850':
            for em in range(NUM_ENSEMBLE_MEMBERS):
                vort.append(self._cvorticity(self.u850[em], self.v850[em]))
                # vort4.append(self._cvorticity4(self.u[em], self.v[em]))
        setattr(self, 'vort{}'.format(pressure_level), vort)

    def _find_min_max_from_fields(self):
        '''Finds the minima (prmsl) and maxima (vort/vort4)'''
        if 'prmsl' in self.fields:
            self.pmins, self.pmaxs = [], []
            for ensemble_member in range(NUM_ENSEMBLE_MEMBERS):
                e, index_pmaxs, index_pmins = cfind_extrema(self.prmsl[ensemble_member])
                self.pmins.append([(self.prmsl[ensemble_member][pmin[0], pmin[1]], (self.lons[pmin[1]], self.lats[pmin[0]]))
                              for pmin in index_pmins])

        if 'u9950' in self.fields and 'v9950' in self.fields:
            self.vmaxs9950 = []
            for ensemble_member in range(NUM_ENSEMBLE_MEMBERS):
                e, index_vmaxs, index_vmins = cfind_extrema(self.vort9950[ensemble_member])
                self.vmaxs9950.append([
                    (self.vort9950[ensemble_member][vmax[0], vmax[1]], (self.lons[vmax[1]], self.lats[vmax[0]]))
                    for vmax in index_vmaxs])

        if 'u850' in self.fields and 'v850' in self.fields:
            self.vmaxs850 = []
            for ensemble_member in range(NUM_ENSEMBLE_MEMBERS):
                e, index_vmaxs, index_vmins = cfind_extrema(self.vort850[ensemble_member])
                self.vmaxs850.append([
                    (self.vort850[ensemble_member][vmax[0], vmax[1]], (self.lons[vmax[1]], self.lats[vmax[0]]))
                    for vmax in index_vmaxs])



import os

# Directories where data and output will be saved.
SETTINGS_DIR = os.path.abspath(os.path.dirname(__file__))
DATA_DIR = os.path.expandvars('$HOME/stormtracks_data/data')
OUTPUT_DIR = os.path.expandvars('$HOME/stormtracks_data/output')
SECOND_OUTPUT_DIR = os.path.expandvars('$HOME/stormtracks_data/output')
LOGGING_DIR = 'logs'
FIGURE_OUTPUT_DIR = os.path.expandvars('$HOME/stormtracks_data/figures')

# 20th C Reanalysis project version.
C20_VERSION = 'v2'

# Lat/Lon range.
MIN_LON = 260
MAX_LON = 340
MIN_LAT = 0
MAX_LAT = 60

RESULTS = 'prod_release_1'

CONSOLE_LOG_LEVEL = 'info'
FILE_LOG_LEVEL = 'debug'

CHUNK_SIZE = 1024*1000
MINIMUM_DOWNLOAD_RATE_1 = 300000 # B/s - 0.3 MB/s.
MINIMUM_DOWNLOAD_RATE_2 = 1000000 # B/s - 1 MB/s.



from collections import OrderedDict, namedtuple
import datetime as dt

import numpy as np
import pandas as pd

from .. import setup_logging
from ..utils.utils import dist, geo_dist, find_extrema
from ..load_settings import settings

log = setup_logging.get_logger('st.find_vortmax')

VortMax = namedtuple('VortMax', ['date', 'pos', 'vort'])

NUM_ENSEMBLE_MEMBERS = 56


class VortmaxFinder(object):
    '''Finds all vortmaxes across ensemble members'''
    def __init__(self, c20data, use_dist_cutoff=True):
        self.c20data = c20data

        # Some settings to document/consider playing with.
        self.use_vort_cutoff = True
        self.use_dist_cutoff = use_dist_cutoff
        self.use_range_cutoff = True
        self.use_geo_dist = True

        if self.use_geo_dist:
            self.dist = geo_dist
            self.dist_cutoff = geo_dist((0, 0), (2, 0)) * 5
        else:
            self.dist = dist
            self.dist_cutoff = 5

        # self.vort_cutoff = 5e-5 # Old value with wrong vort calc.
        # self.vort_cutoff = 2.5e-5
        self.vort_cutoff = 1e-5
	log.info('VortmaxFinder setup:')
	for setting in ['use_vort_cutoff',
		        'use_dist_cutoff',
			'use_range_cutoff',
			'use_geo_dist',
			'vort_cutoff']:
	    log.info('{}: {}'.format(setting, getattr(self, setting)))

    def find_vort_maxima(self, start_date, end_date):
        '''Runs over the date range looking for all vorticity maxima'''
        if start_date < self.c20data.dates[0]:
            raise Exception('Start date is out of date range, try setting the year appropriately')
        elif end_date > self.c20data.dates[-1]:
            raise Exception('End date is out of date range, try setting the year appropriately')
	log.info('finding vortmaxima in range {}-{}'.format(start_date, end_date))
        index = np.where(self.c20data.dates == start_date)[0][0]
        end_index = np.where(self.c20data.dates == end_date)[0][0]

        self.all_vortmax_time_series = []
        results = []

        start = dt.datetime.now()

        for ensemble_member in range(NUM_ENSEMBLE_MEMBERS):
            self.all_vortmax_time_series.append(OrderedDict())

        while index <= end_index:
            date = self.c20data.dates[index]
            self.c20data.set_date(date)

            print('Finding vortmaxima: {0}'.format(date))
            log.debug('Finding vortmaxima: {0}'.format(date))

            for ensemble_member in range(NUM_ENSEMBLE_MEMBERS):
                vortmax_time_series = self.all_vortmax_time_series[ensemble_member]
                vortmaxes = []
                vmaxs = self.c20data.vmaxs850[ensemble_member]

                for vmax in vmaxs:
                    if (self.use_range_cutoff and not
			(settings.MIN_LON <= vmax[1][0] <= settings.MAX_LON and
			 settings.MIN_LAT <= vmax[1][1] <= settings.MAX_LAT)):
                        continue

                    if self.use_vort_cutoff and vmax[0] < self.vort_cutoff:
                        continue

                    vortmax = VortMax(date, vmax[1], vmax[0])
                    vortmaxes.append(vortmax)

                if self.use_dist_cutoff:
                    secondary_vortmaxes = []
                    for i in range(len(vortmaxes)):
                        v1 = vortmaxes[i]
                        for j in range(i + 1, len(vortmaxes)):
                            v2 = vortmaxes[j]
                            if self.dist(v1.pos, v2.pos) < self.dist_cutoff:
                                if v1.vort > v2.vort:
                                    v1.secondary_vortmax.append(v2)
                                    secondary_vortmaxes.append(v2)
                                elif v1.vort <= v2.vort:
                                    v2.secondary_vortmax.append(v1)
                                    secondary_vortmaxes.append(v1)

                    for v in secondary_vortmaxes:
                        if v in vortmaxes:
                            vortmaxes.remove(v)

                vortmax_time_series[date] = vortmaxes

                for vortmax in vortmaxes:
                    row = {'date': date,
                           'em': ensemble_member,
                           'lon': vortmax.pos[0],
                           'lat': vortmax.pos[1],
                           'vort850': vortmax.vort}
                    res = self.get_other_fields(ensemble_member, vortmax)
                    row.update(res)
                    results.append(row)

            index += 1

	end = dt.datetime.now()
	log.info('Found vortmaxima and fields in {}'.format(end - start))

        columns = ['date', 'em', 
                   'lon', 
                   'lat', 
                   'max_ws_lon',
                   'max_ws_lat',
                   'pmin_lon',
                   'pmin_lat',
                   'vort9950', 
                   'vort850', 
                   'max_ws',
                   'prmsl',
                   'pmin_dist',
                   'pmin',
                   'p_ambient_diff',
                   't850',
                   't9950',
                   'cape',
                   'pwat']
        df = pd.DataFrame(results, columns=columns)
        end = dt.datetime.now()
        return df

    def get_other_fields(self, ensemble_member, vortmax):
        res = {}
        vmax_pos = vortmax.pos
        # Round values to the nearest multiple of 2 (vmax_pos can come from an interpolated field)
        # vmax_pos = tuple([int(round(p / 2.)) * 2 for p in actual_vmax_pos])
        lon_index = np.where(self.c20data.lons == vmax_pos[0])[0][0]
        lat_index = np.where(self.c20data.lats == vmax_pos[1])[0][0]

        min_lon, max_lon = lon_index - 5, lon_index + 6
        min_lat, max_lat = lat_index - 5, lat_index + 6
        local_slice = (slice(min_lat, max_lat), slice(min_lon, max_lon))

        local_prmsl = self.c20data.prmsl[ensemble_member][local_slice].copy()

        local_windspeed = np.sqrt(self.c20data.u9950[ensemble_member][local_slice] ** 2 +
                                  self.c20data.v9950[ensemble_member][local_slice] ** 2)
        max_windspeed_pos = np.unravel_index(np.argmax(local_windspeed), (11, 11))
        max_windspeed = local_windspeed[max_windspeed_pos]

        lon = self.c20data.lons[min_lon + max_windspeed_pos[1]]
        lat = self.c20data.lats[min_lat + max_windspeed_pos[0]]

        res['max_ws'] = max_windspeed
        res['max_ws_lon'] = lon
        res['max_ws_lat'] = lat

        res['prmsl'] = self.c20data.prmsl[ensemble_member][lat_index, lon_index]
        e, index_pmaxs, index_pmins = find_extrema(local_prmsl)
        min_dist = 1000
        pmin = None
        pmin_pos = None
        for index_pmin in index_pmins:
            lon = self.c20data.lons[min_lon + index_pmin[1]]
            lat = self.c20data.lats[min_lat + index_pmin[0]]

            local_pmin = local_prmsl[index_pmin[0], index_pmin[1]]
            local_pmin_pos = (lon, lat)
            dist = geo_dist(vmax_pos, local_pmin_pos)
            if dist < min_dist:
                min_dist = dist
                pmin = local_pmin
                pmin_pos = local_pmin_pos

        res['pmin_dist'] = min_dist
        if pmin:
            res['pmin'] = pmin
            res['pmin_lon'] = pmin_pos[0]
            res['pmin_lat'] = pmin_pos[1]
            res['p_ambient_diff'] = local_prmsl.mean() - pmin
        else:
            res['pmin'] = local_prmsl.min()
            res['pmin_lon'] = None
            res['pmin_lat'] = None
            res['p_ambient_diff'] = local_prmsl.mean() - local_prmsl.min()

        res['vort9950'] = self.c20data.vort9950[ensemble_member][lat_index, lon_index]
        res['t850'] = self.c20data.t850[ensemble_member][lat_index, lon_index]
        res['t9950'] = self.c20data.t9950[ensemble_member][lat_index, lon_index]
        res['cape'] = self.c20data.cape[ensemble_member][lat_index, lon_index]
        res['pwat'] = self.c20data.pwat[ensemble_member][lat_index, lon_index]
        # No longer using due to it not having much
        # discriminatory power and space constraints.
        # cyclone_track.rh995s[date] = 0
        return res



import math
import numpy as np
import pylab as plt


def contains(array, i, j, val, side):
    if side == 'left':
        if j == 0:
            return False
        jj = j - 1
    if side == 'right':
        if j == array.shape[1] - 1:
            return False
        jj = j + 1

    r = range(max(i - 1, 0), min(i + 2, array.shape[0]))
    if val == 2:
        r = r[::-1]

    for ii in r:
        if array[ii, jj] == val:
            return True
    return False


def left_contains(array, i, j, val):
    return contains(array, i, j, val, 'left')


def right_contains(array, i, j, val):
    return contains(array, i, j, val, 'right')


def fill_raster(array):
    'Takes an outline of a shape (raster), and fills in the central cells'
    # Not working for cases outline is concave in x-dir.
    filled_a = np.copy(array)
    working_a = np.zeros_like(array)

    tops = []
    bottoms = []
    sides = []
    for j in range(array.shape[1]):
        is_top = True
        top = None
        bottom = None
        for i in range(array.shape[0]):
            if array[i, j]:
                if is_top:
                    is_top = False
                    top = [i, j]
                    bottom = [i, j]
                else:
                    bottom = [i, j]
        if top:
            working_a[top[0], top[1]] = 4
            working_a[bottom[0], bottom[1]] = 2
            tops.append(top)
            bottoms.append(bottom)

    for top, bottom in zip(tops, bottoms):
        found_bottom = False
        while top != bottom:
            top[0] += 1
            filled_a[top[0], top[1]] = 1

    return filled_a, working_a


def path_to_raster(path):
    'Takes a path in with path[:, 1] as the x vals and returns a raster outline'
    # Not working for cases when the path stretches over multiple grid cells.
    imin = int(math.floor(path[:, 1].min()))
    imax = int(math.floor(path[:, 1].max()) + 1)
    jmin = int(math.floor(path[:, 0].min()))
    jmax = int(math.floor(path[:, 0].max()) + 1)
    a = np.zeros((imax - imin, jmax - jmin))
    for point in path:
        on_x_boundary = False
        on_y_boundary = False

        if not point[0].is_integer():
            on_x_boundary = True
        elif not point[1].is_integer():
            on_y_boundary = True

        i = int(math.floor(point[1])) - imin
        j = int(math.floor(point[0])) - jmin
        a[i, j] = 1
        if on_x_boundary and i > 0:
            a[i - 1, j] = 1
        if on_y_boundary and j > 0:
            a[i, j - 1] = 1
    return a



from collections import OrderedDict, namedtuple
import datetime as dt

import numpy as np

from .. import setup_logging
from ..utils.utils import dist, geo_dist, pairwise, find_extrema

log = setup_logging.get_logger('st.tracking')

NUM_ENSEMBLE_MEMBERS = 56


class VortmaxNearestNeighbourTracker(object):
    '''Simple nearest neighbour tracker

    Assumes that the two nearest points from one timestep to another belong to the same track.
    '''
    def __init__(self):
        self.use_geo_dist = True

        if self.use_geo_dist:
            self.dist = geo_dist
            self.dist_cutoff = geo_dist((0, 0), (2, 0)) * 8
        else:
            self.dist = dist
            self.dist_cutoff = 5

    def _construct_vortmax_tracks_by_date(self, all_vortmax_time_series):
        self.all_vort_tracks_by_date = []
        for ensemble_member in range(NUM_ENSEMBLE_MEMBERS):
            vortmax_time_series = all_vortmax_time_series[ensemble_member]
            vortmax_tracks = []
            vort_tracks_by_date = OrderedDict()
            # Find all start vortmaxes (those that have no previous vortmaxes)
            # and use these to generate a track.
            for vortmaxes in vortmax_time_series.values():
                for vortmax in vortmaxes:
                    if len(vortmax.prev_vortmax) == 0:
                        vortmax_track = VortMaxTrack(vortmax, ensemble_member)
                        if len(vortmax_track.vortmaxes) >= 6:
                            vortmax_tracks.append(vortmax_track)

                            for date in vortmax_track.vortmax_by_date.keys():
                                if date not in vort_tracks_by_date:
                                    vort_tracks_by_date[date] = []
                                vort_tracks_by_date[date].append(vortmax_track)

                    # Allows vort_tracks_by_date to be serialized.
                    vortmax.next_vortmax = []
                    vortmax.prev_vortmax = []

            self.all_vort_tracks_by_date.append(vort_tracks_by_date)
        return self.all_vort_tracks_by_date


    def track(self, df):
        all_vortmax_time_series = []
        for ensemble_member in range(NUM_ENSEMBLE_MEMBERS):
            vortmax_time_series = OrderedDict()
            all_vortmax_time_series.append(vortmax_time_series)
            for i in df[df.em == ensemble_member].index:
                row = df.loc[i]
                if row.date not in vortmax_time_series:
                    vortmax_time_series[row.date] = []
                vortmax = VortMax(row.date, (row.lon, row.lat), row.vort)
                vortmax_time_series[vortmax.date].append(vortmax)
        print('Made objects')
        return self.track_vort_maxima(all_vortmax_time_series)


    def track_vort_maxima(self, all_vortmax_time_series):
        '''Uses a generated list of vortmaxes to track them from one timestep to another

        :param vortmax_time_series: dict of vortmaxes
        :returns: self.vort_tracks_by_date (OrderedDict)
        '''
        for ensemble_member in range(NUM_ENSEMBLE_MEMBERS):
            log.info('Tracking vortmaxima: {0}'.format(ensemble_member))
            vortmax_time_series = all_vortmax_time_series[ensemble_member]
            # Loops over 2 lists of vortmaxes
            for vs1, vs2 in pairwise(vortmax_time_series.values()):
                for v1 in vs1:
                    min_dist = self.dist_cutoff
                    v2next = None

                    # Find the nearest vortmax in the next timestep.
                    for v2 in vs2:
                        d = self.dist(v1.pos, v2.pos)
                        if d < min_dist:
                            min_dist = d
                            v2next = v2

                    # Add the nearest vortmax in the next timestep.
                    if v2next:
                        v1.add_next(v2next)
                        if len(v1.next_vortmax) != 1:
                            raise Exception('There should only ever be one next_vortmax')

            # Some vortmaxes may have more than one previous vortmax.
            # Find these and choose the nearest one as the actual previous.
            for vs in vortmax_time_series.values():
                for v in vs:
                    if len(v.prev_vortmax) > 1:
                        min_dist = self.dist_cutoff
                        vprev = None
                        for pv in v.prev_vortmax:
                            d = self.dist(pv.pos, v.pos)
                            if d < min_dist:
                                min_dist = d
                                vprev = pv

                        for pv in v.prev_vortmax:
                            if pv != vprev:
                                pv.next_vortmax.remove(v)

                        v.prev_vortmax = [vprev]

        return self._construct_vortmax_tracks_by_date(all_vortmax_time_series)


class FieldFinder(object):
    def __init__(self, c20data, all_vort_tracks_by_date):
        self.c20data = c20data
        self.all_vort_tracks_by_date = all_vort_tracks_by_date
        self.all_cyclone_tracks = []
        for ensemble_member in range(NUM_ENSEMBLE_MEMBERS):
            self.all_cyclone_tracks.append({})

    def collect_fields(self, start_date, end_date):
        # import ipdb; ipdb.set_trace()
        index = np.where(self.c20data.dates == start_date)[0][0]
        end_index = np.where(self.c20data.dates == end_date)[0][0]

        while index <= end_index:
            date = self.c20data.dates[index]
            log.info('Collecting fields for: {0}'.format(date))
            self.c20data.set_date(date)
            for ensemble_member in range(NUM_ENSEMBLE_MEMBERS):
                cyclone_tracks = self.all_cyclone_tracks[ensemble_member]

                if date not in self.all_vort_tracks_by_date[ensemble_member]:
                    continue

                vort_tracks = self.all_vort_tracks_by_date[ensemble_member][date]
                for vort_track in vort_tracks:
                    cyclone_track = None
                    if vort_track not in cyclone_tracks:
                        if len(vort_track.vortmaxes) >= 6:
                            cyclone_track = CycloneTrack(vort_track, ensemble_member)
                            cyclone_tracks[vort_track] = cyclone_track
                    else:
                        cyclone_track = cyclone_tracks[vort_track]

                    if cyclone_track:
                        self.add_fields_to_track(date, cyclone_track, ensemble_member)
            index += 1

    def add_fields_to_track(self, date, cyclone_track, ensemble_member):
        actual_vmax_pos = cyclone_track.get_vmax_pos(date)
        # Round values to the nearest multiple of 2 (vmax_pos can come from an interpolated field)
        vmax_pos = tuple([int(round(p / 2.)) * 2 for p in actual_vmax_pos])
        lon_index = np.where(self.c20data.lons == vmax_pos[0])[0][0]
        lat_index = np.where(self.c20data.lats == vmax_pos[1])[0][0]

        min_lon, max_lon = lon_index - 5, lon_index + 6
        min_lat, max_lat = lat_index - 5, lat_index + 6
        local_slice = (slice(min_lat, max_lat), slice(min_lon, max_lon))

        local_prmsl = self.c20data.prmsl[ensemble_member][local_slice].copy()

	# 9950 pressure level.
        local_windspeed = np.sqrt(self.c20data.u9950[ensemble_member][local_slice] ** 2 +
                                  self.c20data.v9950[ensemble_member][local_slice] ** 2)
        max_windspeed_pos = np.unravel_index(np.argmax(local_windspeed), (11, 11))
        max_windspeed = local_windspeed[max_windspeed_pos]

        lon = self.c20data.lons[min_lon + max_windspeed_pos[1]]
        lat = self.c20data.lats[min_lat + max_windspeed_pos[0]]

        cyclone_track.max_windspeeds[date] = max_windspeed
        cyclone_track.max_windspeed_positions[date] = (lon, lat)

        e, index_pmaxs, index_pmins = find_extrema(local_prmsl)
        min_dist = 1000
        pmin = None
        pmin_pos = None
        for index_pmin in index_pmins:
            lon = self.c20data.lons[min_lon + index_pmin[1]]
            lat = self.c20data.lats[min_lat + index_pmin[0]]

            local_pmin = local_prmsl[index_pmin[0], index_pmin[1]]
            local_pmin_pos = (lon, lat)
            dist = geo_dist(actual_vmax_pos, local_pmin_pos)
            if dist < min_dist:
                min_dist = dist
                pmin = local_pmin
                pmin_pos = local_pmin_pos

        cyclone_track.min_dists[date] = min_dist
        cyclone_track.pmins[date] = pmin
        cyclone_track.pmin_positions[date] = pmin_pos

        if pmin:
            cyclone_track.p_ambient_diffs[date] = local_prmsl.mean() - pmin
        else:
            cyclone_track.p_ambient_diffs[date] = None

        cyclone_track.t850s[date] = self.c20data.t850[ensemble_member][lat_index, lon_index]
        cyclone_track.t995s[date] = self.c20data.t995[ensemble_member][lat_index, lon_index]
        cyclone_track.capes[date] = self.c20data.cape[ensemble_member][lat_index, lon_index]
        cyclone_track.pwats[date] = self.c20data.pwat[ensemble_member][lat_index, lon_index]
        # No longer using due to it not having much
        # discriminatory power and space constraints.
        cyclone_track.rh995s[date] = 0






import sys
import datetime as dt

import pandas as pd
import numpy as np

# scikit-learn classifiers.
from sklearn.linear_model import SGDClassifier
from sklearn.qda import QDA
from sklearn.lda import LDA
from sklearn.preprocessing import StandardScaler

from stormtracks.results import StormtracksResultsManager, ResultNotFound

# COLS = ['vort', 'max_ws', 'pmin', 't850', 't995', 'cape', 'pwat']
# COLS = ['vort', 'max_ws', 'pmin', 't850', 't995']
# COLS = ['vort', 'max_ws', 'pmin', 't850', 't_anom']
COLS = ['vort850', 'max_ws', 'pmin', 't850']
SGD_LOSSES = ['hinge',
    'modified_huber', 
    'log',
    # 'squared_loss',
    # 'huber',
    # 'epsilon_insensitive'
]
SGD_PENALTY = ['l1', 'l2', 'elasticnet']


def analyse_year(year=2005, results_name='demo'):
    start = dt.datetime.now()

    results_manager = StormtracksResultsManager(results_name)

    try:
        all_fields = results_manager.get_result(year, 'all_fields')
        best_track_matches = results_manager.get_result(year, 'best_track_matches')
    except ResultNotFound:
	print('Could not load all_fields and best_track_matches for {}'.format(year))
	print('Run download_2005.py and process_2005.py first')
	exit(1)

    combined_fields_matches = all_fields.join(best_track_matches)

    # return combined_fields_matches
    analyse_combined(combined_fields_matches)

    end = dt.datetime.now()
    print('Analysed {} in {}'.format(year, end - start))


def analyse_combined(combined_fields_matches):
    data = combined_fields_matches[COLS].values.astype(np.float32)
    are_hurr = ~combined_fields_matches.bt_wind.isnull() & (combined_fields_matches.is_hurr)
    analyse_data(data, are_hurr)


def analyse_data(data, are_hurr):
    sgd = SGDClassifier(loss='hinge', penalty='l1')
    scaler = StandardScaler()
    scaler.fit(data)
    scaled_data = scaler.transform(data)
    fit(sgd, scaled_data, are_hurr)
    predict(sgd, scaled_data, are_hurr)


def predict(classifier, data, are_hurr):
    are_hurr_pred = classifier.predict(data)

    tp = (are_hurr & are_hurr_pred)
    tn = (~are_hurr & ~are_hurr_pred)
    fp = (~are_hurr & are_hurr_pred)
    fn = (are_hurr & ~are_hurr_pred)

    print(', '.join(COLS))

    print('tp: {}'.format(tp.sum()))
    print('tn: {}'.format(tn.sum()))
    print('fp: {}'.format(fp.sum()))
    print('fn: {}'.format(fn.sum()))

    sens = 1. * tp.sum() / (tp.sum() + fn.sum())
    ppv = 1. * tp.sum() / (tp.sum() + fp.sum())
    print('sens: {}'.format(sens))
    print('ppv : {}'.format(ppv))
    print('sens*ppv : {}'.format(sens * ppv))


def fit(classifier, data, are_hurr):
    classifier.fit(data, are_hurr)
    return classifier


if __name__ == '__main__':
    analyse_year()



from __future__ import print_function

import copy
import os
import datetime as dt
from glob import glob

import simplejson
import pylab as plt
import numpy as np
from mpl_toolkits.basemap import Basemap

from ..utils.c_wrapper import cvort, cvort4
from ..load_settings import settings

SAVE_FILE_TPL = 'plot_{0}.json'

DEFAULT_PLOTTER_LAYOUT = {
    'version': 0.1,
    'date': None,
    'name': None,
    'figure': 1,
    'plot_settings': [],
    'ensemble_member': 0,
    'ensemble_mode': 'member',
    }

DEFAULT_PLOT_SETTINGS = {
    'subplot': '111',
    'field': 'prmsl',
    'points': 'pmins',
    'loc': 'wa',
    'vmax': None,
    'vmin': None,
    'best_track_index': -1,
    'vort_max_track_index': -1,
    'pressure_min_track_index': -1,
    'match_index': -1,
    }


def plot_pandas_matched_unmatched(df_year, bt_matches, f1='vort', f2='t850s', flip=False, alpha=0.5):
    plt.clf()
    comb = df_year.join(bt_matches)
    matched = comb[~comb.bt_name.isnull()]
    unmatched = comb[comb.bt_name.isnull()]
    if not flip:
        zo1 = 1
        zo2 = 2
        alpha1=1
        alpha2=alpha
    else:
        zo1 = 2
        zo2 = 1
        alpha1=alpha
        alpha2=1
    uf1 = getattr(unmatched, f1)
    uf2 = getattr(unmatched, f2)
    mf1 = getattr(matched, f1)
    mf2 = getattr(matched, f2)
    plt.scatter(uf1, uf2, c='k', marker='s', zorder=zo1, alpha=alpha1)
    plt.scatter(mf1, mf2, c=matched.bt_wind, marker='s', zorder=zo2, alpha=alpha2)
    plt.xlim((uf1.min(), uf1.max()))
    plt.ylim((uf2.min(), uf2.max()))


def plot_pandas_tracks(ib, matched, mode='all'):
    for bt in ib.best_tracks:
        plt.plot(bt.lons, bt.lats, 'r-', zorder=1)
        plt.scatter(x=bt.lons, y=bt.lats, c=bt.winds, marker='o', zorder=2)
        if mode == 'all':
            lon_lat = matched[matched.bt_name == bt.name]
            plt.scatter(lon_lat.lon, lon_lat.lat, marker='x', zorder=0)
        elif mode == 'mean':
            lon_lat = matched[matched.bt_name == bt.name].groupby('date')['lon', 'lat'].mean()
            lon_lat_std = matched[matched.bt_name == bt.name].groupby('date')['lon', 'lat'].std()
            plt.errorbar(lon_lat.lon, lon_lat.lat, xerr=lon_lat_std.lon, yerr=lon_lat_std.lat, marker='x', zorder=0)
            plt.clim(0, 150)
        elif mode == 'median':
            lon_lat = matched[matched.bt_name == bt.name].groupby('date')['lon', 'lat'].median()
            lon_lat_std = matched[matched.bt_name == bt.name].groupby('date')['lon', 'lat'].std()
            plt.errorbar(lon_lat.lon, lon_lat.lat, xerr=lon_lat_std.lon, yerr=lon_lat_std.lat, marker='x', zorder=0)
            plt.clim(0, 150)
        plt.title('Best track {}: {} - {}'.format(bt.name, bt.dates[0], bt.dates[-1]))

        plt.pause(.1)
        raw_input()
        plt.clf()



def datetime_encode_handler(obj):
    if hasattr(obj, 'isoformat'):
        return {'__isotime__': obj.isoformat()}
    else:
        raise TypeError('Object of type %s with value of {0} is not JSON serializable'.format(
            type(obj), repr(obj)))


def datetime_decode_hook(dct):
    if '__isotime__' in dct:
        return dt.datetime.strptime(dct['__isotime__'], '%Y-%m-%dT%H:%M:%S')
    return dct


class Plotter(object):
    def __init__(self, title, best_tracks, c20data, all_matches):
        self.title = title

        self.best_tracks = best_tracks

        self.c20data = c20data
        self.all_matches = all_matches
        self.date = self.c20data.first_date()

        self.layout = copy.copy(DEFAULT_PLOTTER_LAYOUT)
        self.layout['date'] = self.date
        self.layout['plot_settings'].append(DEFAULT_PLOT_SETTINGS)

        self.date = None
        self.field_names = ['prmsl', 'vort', 'vort4', 'windspeed']
        self.track_names = ['best', 'vort_max', 'pressure_min']
        self.point_names = ['vort_max', 'pressure_min']

    def plot_match_from_best_track(self, best_track):
        for matches in self.all_matches:
            for match in matches:
                if best_track.name == match.best_track.name:
                    for settings in self.layout['plot_settings']:
                        settings['match_index'] = matches.index(match)
                    self.plot()
                    return
        print('Match not found for track')

    def plot(self):
        ensemble_member = self.layout['ensemble_member']
        date = self.layout['date']

        plt.figure(self.layout['figure'])
        plt.clf()
        for plot_settings in self.layout['plot_settings']:
            plt.subplot(plot_settings['subplot'])

            title = [self.title, str(date), str(ensemble_member)]

            if plot_settings['field']:
                title.append(plot_settings['field'])
                field = getattr(self.c20data, plot_settings['field'])
                raster_on_earth(self.c20data.lons, self.c20data.lats, field,
                                plot_settings['vmin'], plot_settings['vmax'], plot_settings['loc'])
            else:
                raster_on_earth(self.c20data.lons, self.c20data.lats,
                                None, None, None, plot_settings['loc'])

            if plot_settings['points']:
                title.append(plot_settings['points'])
                points = getattr(self.c20data, plot_settings['points'])
                for p_val, p_loc in points:
                    plot_point_on_earth(p_loc[0], p_loc[1], 'kx')

            if plot_settings['best_track_index'] != -1:
                plot_ibtrack_with_date(self.best_tracks[plot_settings['best_track_index']], date)

            if plot_settings['vort_max_track_index'] != -1:
                pass

            if plot_settings['pressure_min_track_index'] != -1:
                pass

            if plot_settings['match_index'] != -1:
                plot_match_with_date(
                    self.all_matches[ensemble_member][plot_settings['match_index']], date)

            plt.title(' '.join(title))

    def save(self, name):
        self.layout['name'] = name
        f = open(os.path.join(settings.SETTINGS_DIR, 'plots',
                              SAVE_FILE_TPL.format(self.layout['name'])), 'w')
        simplejson.dump(self.layout, f, default=datetime_encode_handler, indent=4)

    def load(self, name, is_plot=True):
        try:
            f = open(os.path.join(settings.SETTINGS_DIR, 'plots', SAVE_FILE_TPL.format(name)), 'r')
            layout = simplejson.load(f, object_hook=datetime_decode_hook)
            print(layout)
            if layout['version'] != self.layout['version']:
                r = raw_input('version mismatch, may not work! Press c to continue anyway: ')
                if r != 'c':
                    raise Exception('Version mismatch (user cancelled)')

            self.layout['name'] = name
            self.layout = layout
            self.set_date(self.layout['date'], is_plot)
        except Exception, e:
            print('Settings {0} could not be loaded'.format(name))
            print('{0}'.format(e.message))
            raise

    def print_list(self):
        for plot_settings_name in self.list():
            print(plot_settings_name)

    def delete(self, name):
        file_name = os.path.join(settings.SETTINGS_DIR, 'plots', SAVE_FILE_TPL.format(name))
        os.remove(file_name)

    def list(self):
        plot_settings = []
        for fn in glob(os.path.join(settings.SETTINGS_DIR, 'plots', SAVE_FILE_TPL.format('*'))):
            plot_settings.append('_'.join(os.path.basename(fn).split('.')[0].split('_')[1:]))
        return plot_settings

    def set_date(self, date, is_plot=True):
        self.layout['date'] = self.c20data.set_date(
            date, self.layout['ensemble_member'], self.layout['ensemble_mode'])
        if is_plot:
            self.plot()

    def next_date(self):
        self.layout['date'] = self.c20data.next_date(
            self.layout['ensemble_member'], self.layout['ensemble_mode'])
        self.plot()

    def prev_date(self):
        self.layout['date'] = self.c20data.prev_date(
            self.layout['ensemble_member'], self.layout['ensemble_mode'])
        self.plot()

    def set_ensemble_member(self, ensemble_member):
        self.layout['ensemble_member'] = ensemble_member
        self.c20data.set_date(
            self.layout['date'], self.layout['ensemble_member'], self.layout['ensemble_mode'])
        self.plot()

    def next_ensemble_member(self):
        self.set_ensemble_member(self.layout['ensemble_member'] + 1)

    def prev_ensemble_member(self):
        self.set_ensemble_member(self.layout['ensemble_member'] - 1)

    def set_match(self, match_index):
        self.match_index = match_index
        for plot_settings in self.layout['plot_settings']:
            plot_settings.match_index = match_index
        self.plot()

    def next_match(self):
        self.set_match(self.match_index + 1)

    def prev_match(self):
        self.set_match(self.match_index - 1)

    def interactive_plot(self):
        cmd = ''
        args = []
        while True:
            try:
                prev_cmd = cmd
                prev_args = args

                if cmd not in ['c', 'cr']:
                    r = raw_input('# ')

                if r == '':
                    cmd = prev_cmd
                    args = prev_args
                else:
                    try:
                        cmd = r.split(' ')[0]
                        args = r.split(' ')[1:]
                    except:
                        cmd = r
                        args = None

                if cmd == 'q':
                    break
                elif cmd == 'pl':
                    print('plot')
                    self.next_date()
                    self.plot()
                elif cmd == 'n':
                    print('next')
                    self.next_date()
                    self.plot()
                elif cmd == 'p':
                    print('prev')
                    self.prev_date()
                    self.plot()
                elif cmd == 'c':
                    print('continuing')
                    self.c20data.next_date()
                    plt.pause(0.01)
                    self.plot()
                elif cmd == 'cr':
                    print('continuing backwards')
                    self.c20data.prev_date()
                    plt.pause(0.01)
                    self.plot()
                elif cmd == 's':
                    self.save(args[0])
                elif cmd == 'l':
                    self.load(args[0])
                elif cmd == 'ls':
                    self.list()
                elif cmd == 'em':
                    if args[0] == 'n':
                        self.next_ensemble_member()
                    elif args[0] == 'p':
                        self.prev_ensemble_member()
                    else:
                        try:
                            self.set_ensemble_member(int(args[0]))
                        except:
                            pass

            except KeyboardInterrupt, ki:
                # Handle ctrl+c
                # deletes ^C from terminal:
                print('\r', end='')

                cmd = ''
                print('ctrl+c pressed')


def plot_3d_scatter(cyclone_matches, unmatched_cyclones):
    fig = plt.figure(4)
    ax = fig.add_subplot(111, projection='3d')
    ps = {'hu': {'xs': [], 'ys': [], 'zs': []},
          'ts': {'xs': [], 'ys': [], 'zs': []},
          'no': {'xs': [], 'ys': [], 'zs': []}}

    for match in cyclone_matches:
        best_track = match.best_track
        cyclone = match.cyclone
        plotted_dates = []

        for date, cls in zip(best_track.dates, best_track.cls):
            if date in cyclone.dates and cyclone.pmins[date]:
                plotted_dates.append(date)
                if cls == 'HU':
                    ps['hu']['xs'].append(cyclone.pmins[date])
                    ps['hu']['ys'].append(cyclone.p_ambient_diffs[date])
                    ps['hu']['zs'].append(cyclone.vortmax_track.vortmax_by_date[date].vort)
                else:
                    ps['ts']['xs'].append(cyclone.pmins[date])
                    ps['ts']['ys'].append(cyclone.p_ambient_diffs[date])
                    ps['ts']['zs'].append(cyclone.vortmax_track.vortmax_by_date[date].vort)

        for date in cyclone.dates:
            if date not in plotted_dates and cyclone.pmins[date]:
                ps['no']['xs'].append(cyclone.pmins[date])
                ps['no']['ys'].append(cyclone.p_ambient_diffs[date])
                ps['no']['zs'].append(cyclone.vortmax_track.vortmax_by_date[date].vort)

    ax.scatter(c='r', marker='o', **ps['hu'])
    ax.scatter(c='b', marker='o', **ps['ts'])
    ax.scatter(c='b', marker='x', **ps['no'])


SCATTER_ATTRS = {
    # Normally x.
    'vort': {'range': (0, 0.0012)},
    # 1st row.
    'pmin': {'range': (96000, 104000)},
    'pambdiff': {'range': (-1000, 5000)},
    'mindist': {'range': (0, 1000)},
    'max_windspeed': {'range': (0, 100)},
    # 2nd row.
    't995': {'range': (260, 320)},
    't850': {'range': (250, 310)},
    't_anom': {'range': (-15, 10)},
    'max_windspeed_dist': {'range': (0, 1000)},
    'max_windspeed_dir': {'range': (-np.pi, np.pi)},
    'lon': {'range': (200, 360)},
    'lat': {'range': (0, 70)},
}


def plot_2d_scatter(ps, var1, var2, unmatched_lim=None):
    attr1 = SCATTER_ATTRS[var1]
    attr2 = SCATTER_ATTRS[var2]

    plt.xlim(attr1['range'])
    plt.ylim(attr2['range'])

    plt.xlabel(var1)
    plt.ylabel(var2)

    if not unmatched_lim:
        unmatched_lim = len(ps['unmatched']['xs'])

    plt.plot(ps['unmatched']['xs'][:unmatched_lim],
             ps['unmatched']['ys'][:unmatched_lim], 'g+', zorder=0)

    plt.plot(ps['no']['xs'], ps['no']['ys'], 'bx', zorder=1)
    plt.plot(ps['ts']['xs'], ps['ts']['ys'], 'bo', zorder=2)
    plt.plot(ps['hu']['xs'], ps['hu']['ys'], 'ro', zorder=3)


def plot_2d_error_scatter(ps, var1, var2, unmatched_lim=None):
    attr1 = SCATTER_ATTRS[var1]
    attr2 = SCATTER_ATTRS[var2]

    plt.xlim(attr1['range'])
    plt.ylim(attr2['range'])

    plt.xlabel(var1)
    plt.ylabel(var2)

    if not unmatched_lim:
        unmatched_lim = len(ps['un']['xs'])

    plt.plot(ps['un']['xs'][:unmatched_lim], ps['un']['ys'][:unmatched_lim], 'kx', zorder=0)
    plt.plot(ps['fp']['xs'], ps['fp']['ys'], 'bo', zorder=1)
    plt.plot(ps['fn']['xs'], ps['fn']['ys'], 'r^', zorder=2)
    plt.plot(ps['tp']['xs'], ps['tp']['ys'], 'ro', zorder=3)
    plt.plot(ps['tn']['xs'], ps['tn']['ys'], 'b^', zorder=4)


def plot_ensemble_matches(c20data, matches):
    plt.clf()
    raster_on_earth(c20data.lons, c20data.lats, None, None, None, 'wa')
    for match in matches:
        plot_track(match.av_vort_track)
        if match.store_all_tracks:
            for vort_track in match.vort_tracks:
                plot_track(vort_track, 'b+')

        # if raw_input() == 'q':
            # return


def plot_vortmax_track_with_date(vortmax_track, date=None):
    plot_track(vortmax_track, plt_fmt='b--')
    if date:
        try:
            index = np.where(vortmax_track.dates == date)[0][0]
            plot_point_on_earth(vortmax_track.lons[index], vortmax_track.lats[index], 'ko')
        except:
            pass
            # print("Couldn't plot track at date {0} (start {1}, end {2})".format(
            #     date, vortmax_track.dates[0], vortmax_track.dates[-1]))


def plot_ibtrack_with_date(best_track, date=None):
    plot_track(best_track)
    if date:
        try:
            index = np.where(best_track.dates == date)[0][0]
            if best_track.cls[index] == 'HU':
                plot_point_on_earth(best_track.lons[index], best_track.lats[index], 'ro')
            else:
                plot_point_on_earth(best_track.lons[index], best_track.lats[index], 'ko')
        except:
            pass
            # print("Couldn't plot best_track at date {0} (start {1}, end {2})".format(
            #     date, best_track.dates[0], best_track.dates[-1]))


# TODO: integrate with Plotting class.
def time_plot_match(c20data, match):
    best_track = match.best_track
    vort_track = match.vort_track
    for date in best_track.dates:
        c20data.set_date(date)
        plt.clf()

        plt.title(date)

        raster_on_earth(c20data.lons, c20data.lats, c20data.vort, -6e-5, 5e-4, 'wa')
        plot_vortmax_track_with_date(vort_track, date)
        plot_ibtrack_with_date(best_track, date)

        plot_match_dist_with_date(match, date)

        raw_input()


# TODO: integrate with Plotting class.
def time_plot_ibtrack(c20data, best_track):
    for date in best_track.dates:
        c20data.set_date(date)
        plt.clf()

        plt.subplot(2, 1, 1)
        plt.title(date)
        plot_ibtrack_with_date(best_track, date)
        raster_on_earth(c20data.lons, c20data.lats, c20data.prmsl, vmin=99500, vmax=103000, loc='wa')

        plt.subplot(2, 1, 2)
        plot_ibtrack_with_date(best_track, date)
        raster_on_earth(c20data.lons, c20data.lats, c20data.vort, -6e-5, 5e-4, loc='wa')

        raw_input()


# TODO: integrate with Plotting class.
def time_plot_vmax(gdata):
    for date in gdata.vortmax_time_series.keys():
        vortmaxes = gdata.vortmax_time_series[date]
        plt.clf()
        plt.title(date)
        raster_on_earth(gdata.c20data.lons, gdata.c20data.lats, None, 0, 0, 'wa')
        for vmax in vortmaxes:
            plot_point_on_earth(vmax.pos[0], vmax.pos[1], 'ko')
        raw_input()


# TODO: integrate with Plotting class.
def plot_matches(c20data, matches, clear=False):
    for match in matches:
        if clear:
            plt.clf()
            raster_on_earth(c20data.lons, c20data.lats, None, 0, 0, 'wa')
        plot_match_with_date(match)

        raw_input()


def cvorticity(u, v, dx, dy):
    '''Calculates the (2nd order) vorticity by calling into a c function'''
    vort = np.zeros_like(u)
    cvort(u, v, u.shape[0], u.shape[1], dx, dy, vort)
    return vort


def plot_grib_vorticity_at_level(c20gribdata, level_key):
    level = c20gribdata.levels[level_key]
    raster_on_earth(level.lons, level.lats, level.vort, loc='wa')


def plot_match_with_date(match, date=None):
    plot_vortmax_track_with_date(match.vort_track, date)
    plot_ibtrack_with_date(match.best_track, date)
    plot_match_dist_with_date(match, date)
    print('Overlap: {0}, cum dist: {1}, av dist: {2}'.format(
        match.overlap, match.cum_dist, match.av_dist()))


def plot_match_dist_with_date(match, date):
    best_track = match.best_track
    vortmax_track = match.vort_track

    track_index = np.where(best_track.dates == match.overlap_start)[0][0]
    vortmax_track_index = np.where(vortmax_track.dates == match.overlap_start)[0][0]

    while True:
        vortmax = vortmax_track.vortmaxes[vortmax_track_index]
        if date and date == best_track.dates[track_index]:
            plot_path_on_earth(np.array((best_track.lons[track_index], vortmax.pos[0])),
                               np.array((best_track.lats[track_index], vortmax.pos[1])), 'r-')
        else:
            plot_path_on_earth(np.array((best_track.lons[track_index], vortmax.pos[0])),
                               np.array((best_track.lats[track_index], vortmax.pos[1])), 'y-')

        track_index += 1
        vortmax_track_index += 1

        if track_index >= len(best_track.lons) or vortmax_track_index >= len(vortmax_track.lons):
            break


def plot_vort_vort4(c20data, date):
    c20data.set_date(date)
    plt.clf()
    plt.subplot(2, 2, 1)
    plt.title(date)
    raster_on_earth(c20data.lons, c20data.lats, c20data.vort, vmin=-5e-6, vmax=2e-4)

    plt.subplot(2, 2, 3)
    raster_on_earth(c20data.lons, c20data.lats, c20data.vort4, vmin=-5e-6, vmax=2e-4)

    plt.subplot(2, 2, 2)
    raster_on_earth(c20data.lons, c20data.lats, None)
    print(len(c20data.vmaxs))
    for v, vmax in c20data.vmaxs:
        if v > 1e-4:
            plot_point_on_earth(vmax[0], vmax[1], 'go')
        else:
            plot_point_on_earth(vmax[0], vmax[1], 'kx')

    plt.subplot(2, 2, 4)
    raster_on_earth(c20data.lons, c20data.lats, None)

    print(len(c20data.v4maxs))
    for v, vmax in c20data.v4maxs:
        if v > 5e-5:
            plot_point_on_earth(vmax[0], vmax[1], 'go')
        else:
            plot_point_on_earth(vmax[0], vmax[1], 'kx')


def time_plot_ibtracks_pressure_vort(c20data, best_tracks, dates):
    for i, date in enumerate(dates):
        c20data.set_date(date)
        plt.clf()
        plt.subplot(2, 2, 1)
        plt.title(date)
        raster_on_earth(c20data.lons, c20data.lats, c20data.prmsl, vmin=99500, vmax=103000)

        plt.subplot(2, 2, 3)
        raster_on_earth(c20data.lons, c20data.lats, c20data.vort, vmin=-5e-6, vmax=2e-4)

        plt.subplot(2, 2, 2)
        raster_on_earth(c20data.lons, c20data.lats, None)
        for p, pmin in c20data.pmins:
            plot_point_on_earth(pmin[0], pmin[1], 'kx')

        plt.subplot(2, 2, 4)
        raster_on_earth(c20data.lons, c20data.lats, None)

        for v, vmax in c20data.vmaxs:
            if v > 10:
                plot_point_on_earth(vmax[0], vmax[1], 'go')
            else:
                plot_point_on_earth(vmax[0], vmax[1], 'kx')

        for best_track in best_tracks:
            if best_track.dates[0] > date or best_track.dates[-1] < date:
                continue

            index = 0
            for j, track_date in enumerate(best_track.dates):
                if track_date == date:
                    index = j
                    break
            for i in range(4):
                plt.subplot(2, 2, i + 1)
                plot_ibtrack_with_date(best_track, date)

        plt.pause(0.1)
        print(date)
        raw_input()


def time_plot_ibtracks_vort_smoothedvort(c20data, best_tracks, dates):
    if not c20data.smoothing:
        print('Turning on smoothing for c20 data')
        c20data.smoothing = True

    for i, date in enumerate(dates):
        c20data.set_date(date)
        plt.clf()
        plt.subplot(2, 2, 1)
        plt.title(date)
        raster_on_earth(c20data.lons, c20data.lats, c20data.vort)

        plt.subplot(2, 2, 3)
        raster_on_earth(c20data.lons, c20data.lats, c20data.smoothed_vort)

        plt.subplot(2, 2, 2)
        raster_on_earth(c20data.lons, c20data.lats, None)
        for v, vmax in c20data.vmaxs:
            if v > 10:
                plot_point_on_earth(vmax[0], vmax[1], 'go')
            elif v > 4:
                plot_point_on_earth(vmax[0], vmax[1], 'kx')
            else:
                plot_point_on_earth(vmax[0], vmax[1], 'y+')

        plt.subplot(2, 2, 4)
        raster_on_earth(c20data.lons, c20data.lats, None)
        for v, vmax in c20data.smoothed_vmaxs:
            if v > 10:
                plot_point_on_earth(vmax[0], vmax[1], 'go')
            elif v > 4:
                plot_point_on_earth(vmax[0], vmax[1], 'kx')
            else:
                plot_point_on_earth(vmax[0], vmax[1], 'y+')

        for best_track in best_tracks:
            if best_track.dates[0] > date or best_track.dates[-1] < date:
                continue

            index = 0
            for j, track_date in enumerate(best_track.dates):
                if track_date == date:
                    index = j
                    break
            for i in range(4):
                plt.subplot(2, 2, i + 1)
                plot_ibtrack_with_date(best_track, date)

        plt.pause(0.1)
        print(date)
        raw_input()


def plot_ibtracks(best_tracks, start_date, end_date):
    for best_track in best_tracks:
        if best_track.dates[0] >= start_date and best_track.dates[0] <= end_date:
            plot_track(best_track)


def plot_track(track, plt_fmt=None, zorder=1):
    if plt_fmt:
        plot_path_on_earth(track.lons, track.lats, plt_fmt, zorder=zorder)
    else:
        plot_path_on_earth(track.lons, track.lats, 'r-', zorder=zorder)


# TODO: integrate with Plotting class.
def plot_wilma(c20data):
    plot_between_dates(c20data, dt.datetime(2005, 10, 18), dt.datetime(2005, 10, 28))


# TODO: integrate with Plotting class.
def plot_between_dates(c20data, start_date, end_date):
    date = c20data.set_date(start_date)
    while date < end_date:
        plt.clf()

        plt.subplot(2, 2, 1)
        plt.title(date)
        # raster_on_earth(c20data.lons, c20data.lats, c20data.prmsl, 97000, 106000, 'wa')
        raster_on_earth(c20data.lons, c20data.lats, c20data.vort, -6e-5, 5e-4, 'wa')

        plt.subplot(2, 2, 3)
        raster_on_earth(c20data.lons, c20data.lats, c20data.vort4, -6e-5, 5e-4, 'wa')

        plt.subplot(2, 2, 2)
        raster_on_earth(c20data.lons, c20data.lats, None, 0, 0, 'wa')
        for v, vmax in c20data.vmaxs:
            if v > 3e-4:
                plot_point_on_earth(vmax[0], vmax[1], 'ro')
                plt.annotate('{0:2.1f}'.format(v * 1e4), (vmax[0], vmax[1] + 0.2))
            elif v > 2e-4:
                plot_point_on_earth(vmax[0], vmax[1], 'yo')
                plt.annotate('{0:2.1f}'.format(v * 1e4), (vmax[0], vmax[1] + 0.2))
            elif v > 1e-4:
                plot_point_on_earth(vmax[0], vmax[1], 'go')
            else:
                # plot_point_on_earth(vmax[0], vmax[1], 'kx')
                pass

        plt.subplot(2, 2, 4)
        raster_on_earth(c20data.lons, c20data.lats, None, 0, 0, 'wa')
        for v, vmax in c20data.v4maxs:
            if v > 3e-4:
                plot_point_on_earth(vmax[0], vmax[1], 'ro')
                plt.annotate('{0:2.1f}'.format(v * 1e4), (vmax[0], vmax[1] + 0.2))
            elif v > 2e-4:
                plot_point_on_earth(vmax[0], vmax[1], 'yo')
                plt.annotate('{0:2.1f}'.format(v * 1e4), (vmax[0], vmax[1] + 0.2))
            elif v > 1e-4:
                plot_point_on_earth(vmax[0], vmax[1], 'go')
            else:
                # plot_point_on_earth(vmax[0], vmax[1], 'kx')
                pass

        raw_input()
        date = c20data.next_date()
        # plt.pause(0.1)


def lons_convert(lons):
    new_lons = lons.copy()
    m = new_lons > 180
    new_lons[m] = new_lons[m] - 360
    return new_lons


def lon_convert(lon):
    return lon if lon <= 180 else lon - 360


def plot_path_on_earth(lons, lats, plot_fmt=None, zorder=1):
    if plot_fmt:
        plt.plot(lons_convert(lons), lats, plot_fmt, zorder=zorder)
    else:
        plt.plot(lons_convert(lons), lats, zorder=zorder)


def plot_point_on_earth(lon, lat, plot_fmt=None):
    if plot_fmt:
        plt.plot(lon_convert(lon), lat, plot_fmt)
    else:
        plt.plot(lon_convert(lon), lat)


def raster_on_earth(lons, lats, data, vmin=None, vmax=None, loc='earth'):
    if loc == 'earth':
        m = Basemap(projection='cyl', resolution='c',
                    llcrnrlat=-90, urcrnrlat=90, llcrnrlon=-180, urcrnrlon=180)
    elif loc in ['wa', 'west_atlantic']:
        m = Basemap(projection='cyl', resolution='c',
                    llcrnrlat=0, urcrnrlat=60, llcrnrlon=-120, urcrnrlon=-30)
    elif loc in ['nwp', 'northwest_pacific']:
        m = Basemap(projection='cyl', resolution='c',
                    llcrnrlat=0, urcrnrlon=50, llcrnrlon=100, urcrnrlon=180)

    if data is not None:
        plot_lons, plot_data = extend_data(lons, lats, data)
        lons, lats = np.meshgrid(plot_lons, lats)
        x, y = m(lons, lats)
        if vmin:
            m.pcolormesh(x, y, plot_data, vmin=vmin, vmax=vmax)
        else:
            m.pcolormesh(x, y, plot_data)

    m.drawcoastlines()

    p_labels = [0, 1, 0, 0]

    m.drawparallels(np.arange(-90., 90.1, 45.), labels=p_labels, fontsize=10)
    m.drawmeridians(np.arange(-180., 180., 60.), labels=[0, 0, 0, 1], fontsize=10)


def extend_data(lons, lats, data):
    if False:
        # TODO: probably doesn't work!
        # Adds extra data at the end.
        plot_offset = 2
        plot_lons = np.zeros((lons.shape[0] + plot_offset,))
        plot_lons[:-plot_offset] = lons
        plot_lons[-plot_offset:] = lons[-plot_offset:] + 3.75 * plot_offset

        plot_data = np.zeros((data.shape[0], data.shape[1] + plot_offset))
        plot_data[:, :-plot_offset] = data
        plot_data[:, -plot_offset:] = data[:, :plot_offset]
    else:
        # Adds extra data before the start.
        delta = lons[1] - lons[0]
        plot_offset = 180
        plot_lons = np.ma.zeros((lons.shape[0] + plot_offset,))
        plot_lons[plot_offset:] = lons
        plot_lons[:plot_offset] = lons[-plot_offset:] - delta * (lons.shape[0])

        plot_data = np.ma.zeros((data.shape[0], data.shape[1] + plot_offset))
        plot_data[:, plot_offset:] = data
        plot_data[:, :plot_offset] = data[:, -plot_offset:]

    return plot_lons, plot_data



#!/usr/bin/env python
import os
try:
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup, Extension

from stormtracks.version import get_version


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='stormtracks',
    version=get_version(),
    description='Tropical Cyclone Detection and Tracking',
    long_description=read('README.rst'),
    author='Mark Muetzelfeldt',
    author_email='markmuetz@gmail.com',
    maintainer='Mark Muetzelfeldt',
    maintainer_email='markmuetz@gmail.com',

    packages=['stormtracks',
              'stormtracks.utils', 
              'stormtracks.processing',
              'stormtracks.analysis',
              'stormtracks.installation',
              'stormtracks.installation.scripts',
              'stormtracks.installation.settings'],
    scripts=[
        'stormtracks/stormtracks-admin.py',
        ],

    ext_modules=[Extension('stormtracks', ['src/cvort.c', 'src/cextrema.c'])],
    install_requires=[
        'pip',
        'ipython',
        'termcolor',
        'argh',
        ],
    package_data={'stormtracks.installation': [
        'requirements/*.txt', 
        'classifiers/*.json', 
        'plots/*.json'
        ]},
    url='https://github.com/markmuetz/stormtracks',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: C',
        'Topic :: Scientific/Engineering :: Atmospheric Science',
        ],
    keywords=['tropical cyclone track detection'],
    )



import sys
import datetime as dt

import pandas as pd
import numpy as np
import argh

# scikit-learn classifiers.
from sklearn.linear_model import SGDClassifier
from sklearn.qda import QDA
from sklearn.lda import LDA
from sklearn.preprocessing import StandardScaler

from stormtracks.load_settings import settings
from stormtracks.results import StormtracksResultsManager, ResultNotFound

COLS = ['vort850', 'vort9950', 'max_ws', 'pmin', 't850', 't9950', 'cape', 'pwat', 'prmsl', 't_anom']
SGD_LOSSES = ['hinge',
    'modified_huber', 
    'log',
    # These don't work anyway.
    # 'squared_loss',
    # 'huber',
    # 'epsilon_insensitive'
]
SGD_PENALTY = ['l1', 'l2', 'elasticnet']


def cal_val_analysis(cols=None):
    '''Calibrates/validates classifiers from 1990 to 2009

    Calibrates on even years, validates on odd
    Trains all variants of the scikit.learn SGD classifier, as well as LDA/QDA.
    SGD pre-scales data
    Calc's FP/FN/TP/TN + sens, ppv and sens*ppv.
    '''
    # Load the data.
    if cols == None:
        cols = COLS

    cal_years = range(1990, 2010, 2)
    val_years = range(1991, 2010, 2)

    cal_cfm = get_results(cal_years, settings.RESULTS)
    val_cfm = get_results(val_years, settings.RESULTS)

    # Set up classifers.
    classifiers = []
    scalers = []
    for sgd_loss in SGD_LOSSES:
        for sgd_penalty in SGD_PENALTY:
            sgd = SGDClassifier(loss=sgd_loss, penalty=sgd_penalty)
            sgd_scaler = StandardScaler()
            classifiers.append(sgd)
            scalers.append(sgd_scaler)

    classifiers.append(LDA())
    scalers.append(None)

    classifiers.append(QDA())
    scalers.append(None)

    # Perform classification.
    for classifier, scaler in zip(classifiers, scalers):
        print('Analysing with classifier {}'.format(classifier))

        try:
            for cfm, is_cal in [(cal_cfm, True), (val_cfm, False)]:
                if is_cal:
                    print('CAL')
                else:
                    print('VAL')

                data = cfm[cols].values.astype(np.float32)
                are_hurr = ~cfm.bt_wind.isnull() & (cfm.is_hurr)

                if is_cal:
                    if scaler is not None:
                        scaler.fit(data)
                        scaled_data = scaler.transform(data)
                    else:
                        scaled_data = data

                    fit(classifier, scaled_data, are_hurr)
                else:
                    if scaler is not None:
                        scaled_data = scaler.transform(data)
                    else:
                        scaled_data = data

                print(', '.join(cols))
                predict(classifier, scaled_data, are_hurr)
            print('')
        except Exception, e:
            print('Error with classifier {}'.format(classifier))
            print(e)


def get_results(year_range, results_name):
    '''Gets the results for the given years (list)'''
    results_manager = StormtracksResultsManager(results_name)

    max_index = 0
    fields = []
    best_tracks = []
    for year in year_range:
        print(year)

        all_fields = results_manager.get_result(year, 'all_fields')
        all_fields['t_anom'] = all_fields.t9950 - all_fields.t850
        best_track_matches = results_manager.get_result(year, 'best_track_matches')

        all_fields.index += max_index
        best_track_matches.index += max_index
        max_index = all_fields.index.max() + 1

        fields.append(all_fields)
        best_tracks.append(best_track_matches)

    print('Concatenating')
    fields = pd.concat(fields)
    best_tracks = pd.concat(best_tracks)

    print('Joining')
    combined_fields_matches = fields.join(best_tracks)
    return combined_fields_matches


def predict(classifier, data, are_hurr):
    '''Predicts classes for classifier using data'''
    are_hurr_pred = classifier.predict(data)

    tp = (are_hurr & are_hurr_pred)
    tn = (~are_hurr & ~are_hurr_pred)
    fp = (~are_hurr & are_hurr_pred)
    fn = (are_hurr & ~are_hurr_pred)

    print('tp: {}'.format(tp.sum()))
    print('tn: {}'.format(tn.sum()))
    print('fp: {}'.format(fp.sum()))
    print('fn: {}'.format(fn.sum()))

    sens = 1. * tp.sum() / (tp.sum() + fn.sum())
    ppv = 1. * tp.sum() / (tp.sum() + fp.sum())
    print('sens: {}'.format(sens))
    print('ppv : {}'.format(ppv))
    print('sens*ppv : {}'.format(sens * ppv))


def fit(classifier, data, are_hurr):
    '''Fits classifier to the data'''
    classifier.fit(data, are_hurr)


if __name__ == '__main__':
    argh.dispatch_command(cal_val_analysis)
